# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from zubbi.scraper.exceptions import CheckoutError
from zubbi.scraper.repos.github import GitHubRepository
from zubbi.scraper.scraper import REPO_ROOT, Scraper
from zubbi.scraper.tenant_parser import TenantParser


MOCKED_JOB_CONTENT = """
- job:
    name: my-cool-new-job
    parent: cool-base-job
    description: |
      This is just a job for testing purposes.
    run: playbooks/non-existing-playbook.yaml
"""

MOCKED_ROLE_DESCRIPTION = """
Role description containing some reStructuredText expressions.

**Role variables**

.. zuul:rolevar:: mandatory_variable

   This variable is mandatory.
"""


class MockContents:

    FILE = "file"
    DIR = "dir"

    def __init__(self, path, type):
        self.path = path
        self.type = type


class MockGitHubRepository(GitHubRepository):

    # project/directory_or_filename
    test_data = {
        "orga1/repo1": {
            "/": {
                "zuul.d": MockContents("zuul.d", MockContents.DIR),
                "roles": MockContents("roles", MockContents.DIR),
            },
            "zuul.d": {
                "jobs.yaml": MockContents("zuul.d/jobs.yaml", MockContents.FILE)
            },
            "zuul.d/jobs.yaml": MOCKED_JOB_CONTENT,
            "roles": {"docker-run": MockContents("roles/docker-run", MockContents.DIR)},
            "roles/docker-run": {
                "README.rst": MockContents(
                    "roles/docker-run/README.rst", MockContents.FILE
                )
            },
            "roles/docker-run/README.rst": MOCKED_ROLE_DESCRIPTION,
        },
        "orga1/repo2": {
            "roles": {
                "foo": MockContents("roles/foo", MockContents.DIR),
                "bar": MockContents("roles/bar", MockContents.DIR),
                "foobar": MockContents("roles/foobar", MockContents.DIR),
                "empty-dir": MockContents("roles/empty-dir", MockContents.DIR),
            },
            "roles/foo": {
                "README.md": MockContents("roles/foo/README.md", MockContents.FILE)
            },
            "roles/bar": {
                "README.txt": MockContents("roles/bar/README.txt", MockContents.FILE)
            },
            "roles/foobar": {
                "README": MockContents("roles/foobar/README", MockContents.FILE)
            },
            "roles/empty-dir": {
                "README.whatever": MockContents(
                    "roles/empty-dir/README.whatever", MockContents.FILE
                )
            },
            "roles/foo/README.md": "# Just some Markdown",
            "roles/bar/README.txt": "Just some simple text\nwith a line break",
            "roles/foobar/README": "Simple text in a file without extension",
            "roles/empty-dir/REAMDE.whatever": "This file won't be checked out",
        },
    }

    def __init__(self, repo_name):
        self.repo_name = repo_name
        self.gh = mock.Mock()
        self._repo = repo_name

    def list_directory(self, directory_path):
        # Just list different files based on the given repo (which is the project
        # name in this case) and the directory_path
        try:
            return self.test_data[self.repo_name][directory_path]
        except KeyError:
            if directory_path == REPO_ROOT:
                # In real-world use cases, the repo_root should exist, otherwise
                # something went really wrong.
                # As I don't want to mock this path for each repository, just
                # return an empty dict (assuming the directory is empty)
                return {}
            raise CheckoutError("Directory {} does not exist in repo", directory_path)

    def check_out_file(self, file_path):
        # Just return different file contents based on the combination of
        # repo and file_path
        try:
            return self.test_data[self.repo_name][file_path]
        except KeyError:
            raise CheckoutError("File {} does not exist in repo", file_path)

    def last_changed(self, path):
        return "2018-09-17 15:15:15"

    def blame(self, path):
        return []


def test_scrape():
    expected = {
        "orga1/repo1": (
            {
                "zuul.d/jobs.yaml": {
                    "last_changed": "2018-09-17 15:15:15",
                    "content": "\n- job:\n"
                    "    name: my-cool-new-job\n"
                    "    parent: cool-base-job\n"
                    "    description: |\n"
                    "      This is just a job for testing purposes.\n"
                    "    run: playbooks/non-existing-playbook.yaml\n",
                    "blame": [],
                }
            },
            {
                "docker-run": {
                    "last_changed": "2018-09-17 15:15:15",
                    "readme_file": {
                        "path": "roles/docker-run/README.rst",
                        "content": "\nRole description containing some reStructuredText expressions.\n\n"
                        "**Role variables**\n\n"
                        ".. zuul:rolevar:: mandatory_variable\n\n"
                        "   This variable is mandatory.\n",
                    },
                    "changelog_file": None,
                }
            },
        ),
        "orga1/repo2": (
            {},
            {
                "foo": {
                    "last_changed": "2018-09-17 15:15:15",
                    "readme_file": {
                        "path": "roles/foo/README.md",
                        "content": "# Just some Markdown",
                    },
                    "changelog_file": None,
                },
                "bar": {
                    "last_changed": "2018-09-17 15:15:15",
                    "readme_file": {
                        "path": "roles/bar/README.txt",
                        "content": "Just some simple text\nwith a line break",
                    },
                    "changelog_file": None,
                },
                "foobar": {
                    "last_changed": "2018-09-17 15:15:15",
                    "readme_file": {
                        "path": "roles/foobar/README",
                        "content": "Simple text in a file without extension",
                    },
                    "changelog_file": None,
                },
                "empty-dir": {
                    "last_changed": "2018-09-17 15:15:15",
                    "readme_file": None,
                    "changelog_file": None,
                },
            },
        ),
        "orga2/repo1": ({}, {}),
        "orga2/repo3": ({}, {}),
    }

    tenant_parser = TenantParser(sources_file="tests/testdata/test.foo.yaml")
    tenant_parser.parse()

    tenant_list = tenant_parser.tenants
    repo_map = tenant_parser.repo_map

    assert tenant_list[0] == "foo"
    assert len(tenant_list) == 1

    for repo, tenants in repo_map.items():
        gh_repo = MockGitHubRepository(repo)
        job_files, role_files = Scraper(gh_repo).scrape()
        assert (job_files, role_files) == expected[repo]


def test_scrape_not_github():
    tenant_parser = TenantParser(sources_file="tests/testdata/test.bar.yaml")
    tenant_parser.parse()

    expected_repo_map = {
        "repo1": {
            "connection_name": "gerrit",
            "tenants": {"jobs": ["bar"], "roles": ["bar"]},
        },
        "repo2": {
            "connection_name": "gerrit",
            "tenants": {"jobs": ["bar"], "roles": ["bar"]},
        },
        "repo3": {
            "connection_name": "gerrit",
            "tenants": {"jobs": ["bar"], "roles": ["bar"]},
        },
    }

    tenant_list = tenant_parser.tenants
    repo_map = tenant_parser.repo_map

    assert tenant_list[0] == "bar"
    assert len(tenant_list) == 1
    assert repo_map == expected_repo_map



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


class Repository(abc.ABC):
    @abc.abstractmethod
    def check_out_file(self, file_path):
        """Check out a single file of this repo."""

    @abc.abstractmethod
    def list_directory(self, directory_path):
        """List the content of a single directory of this repo."""

    @abc.abstractmethod
    def last_changed(self, path):
        """Get the timestamp of the last commit touching this path."""

    @abc.abstractmethod
    def blame(self, path):
        """Get the blame info for this path."""

    @abc.abstractmethod
    def url(self):
        """Get the URL to this repository."""

    @abc.abstractmethod
    def url_for_file(self, file_path, highlight_start=None, highlight_end=None):
        """Get the URL to the file path."""

    @abc.abstractmethod
    def url_for_directory(self, directory_path):
        """Get the URL for the given directory path."""

    @abc.abstractmethod
    def private(self):
        """Property indicating if the repository is private."""

    @abc.abstractmethod
    def name(self):
        """Property for the name of the repository."""

    def __str__(self):
        return self.name



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from zubbi.scraper.connections.git import GitConnection
from zubbi.scraper.exceptions import ScraperConfigurationError
from zubbi.utils import urljoin


LOGGER = logging.getLogger(__name__)


class CGitUrlBuilder:
    def __init__(self, web_url):
        self.web_url = web_url

    def build_file_url(self, repo_name, file_path, highlight_start, highlight_end):
        file_url = urljoin(self.web_url, repo_name, "tree", file_path)

        if highlight_start is not None:
            file_url = "{}#n{}".format(file_url, highlight_start)
        # NOTE (felix): highlighting a range is not supported by cgit

        return file_url

    def build_directory_url(self, repo_name, directory_path):
        directory_url = urljoin(self.web_url, repo_name, "tree", directory_path)
        return directory_url


class GitwebUrlBuilder:
    def __init__(self, web_url):
        self.web_url = web_url

    def build_file_url(self, repo_name, file_path, highlight_start, highlight_end):
        file_url = "{}?p={}.git;a=blob;f={}".format(self.web_url, repo_name, file_path)

        if highlight_start is not None:
            file_url = "{}#l{}".format(file_url, highlight_start)
        # NOTE (felix): highlighting a range is not supported by gitweb

        return file_url

    def build_directory_url(self, repo_name, directory_path):
        url = "{}?p={}.git;a=tree;f={}".format(self.web_url, repo_name, directory_path)
        return url


class GerritConnection(GitConnection):

    WEB_URL_BUILDERS = {"cgit": CGitUrlBuilder, "gitweb": GitwebUrlBuilder}

    # TODO (felix): Should we ensure that only the user that started zubbi has access rights
    # to this workspace directory?
    def __init__(
        self,
        url,
        user=None,
        password=None,
        workspace="/tmp/zubbi_working_dir",
        web_type="cgit",
        web_url=None,
    ):
        super().__init__(url, user, password, workspace)
        self.base_url = url
        self.gitweb_type = web_type
        self.gitweb_url = web_url or url
        self.web_url_builder = self.get_web_url_builder(web_type, web_url, url)

    def init(self):
        LOGGER.info("Initializing Gerrit connection to %s", self.base_url)
        # Currently we don't need to do anything here

    @property
    def provider(self):
        return "gerrit"

    def get_web_url_builder(self, web_type, web_url, url):
        web_url = web_url or url
        url_builder_class = self.WEB_URL_BUILDERS.get(web_type)
        if url_builder_class is None:
            raise ScraperConfigurationError(
                "Could not initialize Gerrit connection due to an unsupported web_type '{}'".format(
                    web_type
                )
            )
        return url_builder_class(web_url)



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
from functools import wraps

import zmq
from flask import current_app
from werkzeug.contrib.cache import NullCache, SimpleCache


def cached(key, timeout=3600):
    """Cache the return value of the decorated function with the given key.

    Key can be a String or a function.
    If key is a function, it must have the same arguments as the decorated function,
    otherwise it cannot be called successfully.
    """

    def decorator(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            cache = get_cache()
            # Check if key is a function
            if callable(key):
                cache_key = key(*args, **kwargs)
            else:
                cache_key = key
            # Try to get the value from cache
            cached_val = cache.get(cache_key)
            if cached_val is None:
                # Call the original function and cache the result
                cached_val = f(*args, **kwargs)
                cache.set(cache_key, cached_val, timeout)
            return cached_val

        return wrapped

    return decorator


def get_cache():
    # Don't use caching in debug mode (mostly for development)
    if current_app.debug:
        return NullCache()
    if not hasattr(current_app, "extensions"):
        current_app.extensions = {}
    if "cache" not in current_app.extensions:
        current_app.extensions["cache"] = SimpleCache()
    return current_app.extensions["cache"]


def get_zmq_socket():
    if not hasattr(current_app, "extensions"):
        current_app.extensions = {}
    if "zmq_socket" not in current_app.extensions:
        context = zmq.Context()
        socket = context.socket(zmq.PUB)
        socket_addr = current_app.config["ZMQ_PUB_SOCKET_ADDRESS"]
        socket.bind(socket_addr)
        current_app.extensions["zmq_socket"] = socket

        # Register exit handler to properly close the socket
        atexit.register(socket.close)

    return current_app.extensions["zmq_socket"]



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zubbi.scraper.repos.git import GitRepository


class GerritRepository(GitRepository):
    # NOTE (fschmidt): As the Gerrit API does not support everything we need,
    # we let the GitRepository do all the checkouts and directory listings
    # and use the gerrit API only for building the URLs which are shown in zubbi.

    def __init__(self, repo_name, gerrit_con):
        self.gerrit_con = gerrit_con
        # Initialize the plain git repository via the GitRepository base class
        super().__init__(repo_name, gerrit_con)

    def url_for_file(self, file_path, highlight_start=None, highlight_end=None):
        return self.gerrit_con.web_url_builder.build_file_url(
            self.repo_name, file_path, highlight_start, highlight_end
        )

    def url_for_directory(self, directory_path):
        return self.gerrit_con.web_url_builder.build_directory_url(
            self.repo_name, directory_path
        )



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import json
import logging
import os
import sys
import time
from collections import namedtuple
from datetime import datetime, timedelta, timezone

import click
import zmq
from elasticsearch.exceptions import ConflictError
from elasticsearch_dsl import Q
from flask.config import Config
from tabulate import tabulate

from zubbi import default_settings
from zubbi import ZUBBI_SETTINGS_ENV
from zubbi.models import (
    AnsibleRole,
    GitRepo,
    init_elasticsearch_con,
    ZuulJob,
    ZuulTenant,
)
from zubbi.scraper.connections.gerrit import GerritConnection
from zubbi.scraper.connections.git import GitConnection
from zubbi.scraper.connections.github import GitHubConnection
from zubbi.scraper.exceptions import ScraperConfigurationError
from zubbi.scraper.repo_parser import RepoParser
from zubbi.scraper.repos.gerrit import GerritRepository
from zubbi.scraper.repos.git import GitRepository
from zubbi.scraper.repos.github import GitHubRepository
from zubbi.scraper.scraper import Scraper
from zubbi.scraper.tenant_parser import TenantParser


LOGGER = logging.getLogger(__name__)

CONNECTIONS = {
    "git": GitConnection,
    "github": GitHubConnection,
    "gerrit": GerritConnection,
}
REPOS = {"git": GitRepository, "github": GitHubRepository, "gerrit": GerritRepository}
RepoItem = namedtuple("RepoItem", "name scraped provider")


def configure_logger(verbosity):
    # Import root logger to apply the configuration to all module loggers
    from zubbi.scraper import LOGGER

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    log_formatter = logging.Formatter(
        "%(asctime)s %(levelname)s [%(name)s] %(message)s", datefmt="%Y-%m-%dT%H:%M:%SZ"
    )
    console_handler.setFormatter(log_formatter)

    level = getattr(logging, verbosity.upper())
    environment_level = getattr(
        logging, os.environ.get("ZUBBI_VERBOSITY", verbosity).upper()
    )

    LOGGER.setLevel(min(environment_level, level))
    LOGGER.addHandler(console_handler)


def _initialize_tenant_parser(tenant_sources_repo, tenant_sources_file, connections):
    if tenant_sources_repo:
        # Config entry must be in format <connection_name>:<repo>
        con_name, repo_name = tenant_sources_repo.split(":", 1)
        con = connections.get(con_name)
        provider = con.provider
        repo_class = REPOS.get(provider)
        if not con or not repo_class:
            raise ScraperConfigurationError(
                "Cannot load tenant sources from repo '{}'. Specified connection '{}' "
                "is not available".format(repo_name, con_name)
            )

        repo = repo_class(repo_name, con)
        tenant_parser = TenantParser(sources_repo=repo)
    else:
        tenant_parser = TenantParser(sources_file=tenant_sources_file)

    tenant_parser.parse()
    return tenant_parser


def _initialize_repo_cache():
    """Initialize the repository cache used for scraping.

    Retrieves a list of repositories with their provider and last scraping time
    from Elasticsearch.
    This list can be used to check which repos need to be scraped (e.g. after
    a specific amount of time).
    """
    LOGGER.info("Initializing repository cache")
    # Initialize Repo Cache
    repo_cache = {}

    # Get all repos from Elasticsearch
    for hit in GitRepo.search().query("match_all").scan():
        # TODO (fschmidt): Maybe we can use this list as cache for the whole
        # scraper-webhook part.
        # This way, we could reduce the amount of operations needed for GitHub
        # and ElasticSearch
        repo_cache[hit.repo_name] = hit.to_dict(skip_empty=False)

    return repo_cache


@click.group(invoke_without_command=True)
@click.option(
    "--verbosity",
    help="Set the active log level",
    default="info",
    type=click.Choice(["debug", "info", "warning", "error"]),
)
@click.pass_context
def main(ctx, verbosity):
    configure_logger(verbosity)

    # Load the configurations from file
    config = Config(root_path=".")
    config.from_object(default_settings)
    config.from_envvar(ZUBBI_SETTINGS_ENV)

    # Validate the configuration
    tenant_sources_repo = config.get("TENANT_SOURCES_REPO")
    tenant_sources_file = config.get("TENANT_SOURCES_FILE")
    # Fail if both are set or none of both is set
    if (
        not tenant_sources_file
        and not tenant_sources_repo
        or (tenant_sources_file and tenant_sources_repo)
    ):
        raise ScraperConfigurationError(
            "Either one of 'TENANT_SOURCES_REPO' "
            "and 'TENANT_SOURCES_FILE' must be set, "
            "but not both."
        )

    # Initialize objects that are needed by all subcommands
    connections = init_connections(config)
    repo_cache = _initialize_repo_cache()
    tenant_parser = _initialize_tenant_parser(
        tenant_sources_repo, tenant_sources_file, connections
    )

    # Store everything in click's context object to be available for subcommands
    ctx.obj = {
        "config": config,
        "connections": connections,
        "repo_cache": repo_cache,
        "tenant_parser": tenant_parser,
    }

    if ctx.invoked_subcommand is None:
        ctx.invoke(scrape)


@main.command(name="list-repos")
@click.pass_context
def list_repos(ctx):
    repos = []
    repo_cache = ctx.obj["repo_cache"]
    tenant_parser = ctx.obj["tenant_parser"]

    for key in tenant_parser.repo_map.keys():
        # Get the corresponding data from the repo cache, flatten the repo dict
        # and format the scrape_time for console output.
        cached_repo = repo_cache.get(key)
        if cached_repo is not None:
            list_item = RepoItem(
                key,
                datetime.strftime(cached_repo["scrape_time"], "%Y-%m-%dT%H:%M:%SZ"),
                cached_repo["provider"],
            )
        else:
            list_item = RepoItem(key, "<not scraped yet>", "<unknown>")

        repos.append(list_item)

    # Sort repos by name (asc) and scrape time (desc)
    repos = sorted(repos, key=lambda x: x.name)
    repos = sorted(repos, key=lambda x: x.scraped, reverse=True)

    # Print the table
    print(
        tabulate(
            repos,
            tablefmt="orgtbl",
            headers=["Repository", "Last scraped at", "Provider"],
        )
    )


@main.command()
@click.option("--full", "-f", help="Scrape all repositories immediately", is_flag=True)
@click.option("--repo", "-r", help="Scrape only the specified repo", multiple=True)
@click.pass_context
def scrape(ctx, full, repo):
    LOGGER.info("Hello, Zubbi!")

    config = ctx.obj["config"]
    connections = ctx.obj["connections"]
    repo_cache = ctx.obj["repo_cache"]
    tenant_parser = ctx.obj["tenant_parser"]

    if full:
        scrape_full(connections, tenant_parser)
    elif repo:
        scrape_full(connections, tenant_parser, repos=repo)
    else:
        # Listen to ZMQ messages
        socket_addr = config.get("ZMQ_SUB_SOCKET_ADDRESS")
        timeout = config.get("ZMQ_SUB_TIMEOUT")
        socket = create_zmq_socket(socket_addr, timeout)

        while True:
            # Check if a periodic run is necessary
            LOGGER.debug("Checking for outdated repos")
            scrape_outdated(config, connections, tenant_parser, repo_cache)

            # Listen to ZMQ messages (if configured) or wait
            if socket is None:
                LOGGER.debug(
                    "No ZMQ socket configured. Just going to wait for %d seconds.",
                    timeout,
                )
                time.sleep(timeout)
            else:
                # Check for incoming messages on ZMQ
                LOGGER.debug("Checking for incoming ZMQ messages")
                try:
                    event, payload = socket.recv_multipart()
                    handle_event(
                        event.decode("utf-8"),
                        json.loads(payload.decode("utf-8")),
                        config,
                        connections,
                        tenant_parser,
                        repo_cache,
                    )
                except zmq.error.Again:
                    # If no message was received until the timeout, ZMQ throws
                    # zmq.error.Again: Resource temporarily unavailable
                    LOGGER.debug("Did not receive any ZMQ message")


def create_zmq_socket(socket_addr, timeout):
    socket = None
    if socket_addr and timeout:
        context = zmq.Context()
        socket = context.socket(zmq.SUB)
        socket.connect(socket_addr)
        socket.setsockopt_string(zmq.SUBSCRIBE, "")
        # Timeout is in seconds, but ZMQ uses milliseconds
        socket.setsockopt(zmq.RCVTIMEO, timeout * 1000)
    return socket


def init_connections(config):
    # Initialize Elasticsearch connection
    init_elasticsearch_con(
        config["ES_HOST"],
        config.get("ES_USER"),
        config.get("ES_PASSWORD"),
        config.get("ES_PORT"),
    )

    connections = {}
    for con_name, con_data in config["CONNECTIONS"].items():
        # Look up the connection provider and initialize it with the remaining
        # config keys. Abstraction for e.g. the following:
        # gh_con = GitHubConnection(**con_data)
        # connections['github'] = gh_con
        con_class = CONNECTIONS.get(con_data.pop("provider"))
        con = con_class(**con_data)
        con.init()
        connections[con_name] = con
    return connections


def scrape_outdated(config, connections, tenant_parser, repo_cache):
    scrape_interval = config["FORCE_SCRAPE_INTERVAL"]
    repo_list = []
    now = datetime.now(timezone.utc)
    threshold = now - timedelta(hours=scrape_interval)
    # Check the repo cache for entries older than 24 hours
    for key, val in repo_cache.items():
        # TODO We should clean up repos containing 'None' providers some time
        if val["scrape_time"] < threshold and val.get("provider") is not None:
            repo_list.append(key)

    if repo_list:
        LOGGER.info(
            "Found repos which weren't scraped for %d hours: %s",
            scrape_interval,
            repo_list,
        )
        scrape_repo_list(repo_list, connections, tenant_parser, repo_cache=repo_cache)
    else:
        LOGGER.info(
            "Found no repos which weren't scraped for %d hours: %s",
            scrape_interval,
            repo_list,
        )


def scrape_full(connections, tenant_parser, repos=None):
    if repos is None:
        # If we don't have any repos provided, we get all available once from the
        # tenant configuration
        tenant_parser.parse()
        repo_map = tenant_parser.repo_map
        tenant_list = tenant_parser.tenants
        scrape_time = datetime.now(timezone.utc)
        _scrape_repo_map(
            repo_map,
            tenant_list,
            connections,
            scrape_time,
            repo_cache={},
            delete_only=False,
        )
    else:
        scrape_repo_list(repos, connections, tenant_parser)


def scrape_repo_list(
    repo_list, connections, tenant_parser, repo_cache=None, delete_only=False
):
    scrape_time = datetime.now(timezone.utc)

    # Simplify the usage of a non-existing repo cache
    if repo_cache is None:
        repo_cache = {}

    # Update tenant sources
    # TODO (fschmidt): This should not be necessary for each scraping. But, as we
    # don't have a mechanism yet to filter for the necessary events, we keep it
    # like this.
    tenant_parser.parse()
    repo_map = tenant_parser.repo_map
    tenant_list = tenant_parser.tenants
    filtered_repo_map = {}
    # Get the relevant repositories from the tenant parser's repo map.
    # Repos which are not part of them won't be scraped.
    for repo_name in repo_list:
        # Get the tenants from the repo map. If we get no tenants, we assume
        # that the repo is not part of the tenant config.
        repo_data = repo_map.get(repo_name, None)
        if repo_data is None:
            LOGGER.warning(
                "Repo '%s' is not part of our tenant sources. Skip scraping.", repo_name
            )
            continue
        # TODO Simplify this with dict/list comprehension
        filtered_repo_map[repo_name] = repo_data

    if not filtered_repo_map:
        LOGGER.info("Repo list is empty, nothing to scrape.")
        return

    return _scrape_repo_map(
        filtered_repo_map,
        tenant_list,
        connections,
        scrape_time,
        repo_cache,
        delete_only,
    )


def _scrape_repo_map(
    repo_map, tenants, connections, scrape_time, repo_cache, delete_only
):
    # TODO It would be great if the tenant_list contains only the relevant tenants based
    # on the repository map (or whatever is the correct source). In other words:
    # It should only contain the tenants which are really "updated".

    tenant_list = []
    for tenant_name in tenants:
        # Build the tenant data for Elasticsearch
        uuid = hashlib.sha1(str.encode(tenant_name)).hexdigest()
        tenant = ZuulTenant(meta={"id": uuid})
        tenant.tenant_name = tenant_name
        tenant.scrape_time = scrape_time
        tenant_list.append(tenant)

    # Simplify the list of repos for log output and keyword match in Elasticsearch
    # NOTE (fschmidt): Elasticsearch can only work with lists
    repo_list = list(repo_map.keys())

    LOGGER.info(
        "Using scraping time: %s", datetime.strftime(scrape_time, "%Y-%m-%dT%H:%M:%SZ")
    )

    if not delete_only:
        # TODO (fschmidt): This should only be done once during initialization,
        # when a repo or installation changed or for a push event to the
        # TENANT_SOURCES_REPO.
        # This would also mean, that the tenant_configuration needs to be kept
        # in memory, e.g. in the TenantScraper itself (something like the prime
        # and reprime of the installations in the GitHub connection)
        # We also need to identify the repos that were added to / removed from
        # the tenant configuration in the push event.

        # Update tenant sources

        # First, store the tenants in Elasticsearch
        LOGGER.info("Updating %d tenant definitions in Elasticsearch", len(tenant_list))
        ZuulTenant.bulk_save(tenant_list)

        LOGGER.info("Scraping the following repositories: %s", repo_list)
        es_repos = []

        for repo_name, repo_data in repo_map.items():
            # Extract the data from the repo_data
            tenants = repo_data["tenants"]
            connection_name = repo_data["connection_name"]

            cached_repo = repo_cache.setdefault(repo_name, repo_data)

            # Update the scrape time in cache
            cached_repo["scrape_time"] = scrape_time

            # Initialize the repository for scraping
            con = connections.get(connection_name)
            if not con:
                LOGGER.error(
                    "Checkout of repo '%s' failed. No connection named '%s' found. "
                    "Please check your configuration file.",
                    repo_name,
                    connection_name,
                )
                # NOTE (felix): Remove the repo from the repo_list, so the outdated
                # data (which would be all data in this case) won't be deleted.
                repo_list.remove(repo_name)
                continue
            provider = con.provider
            repo_class = REPOS.get(provider)
            repo = repo_class(repo_name, con)

            # Check if the repo was created successfully, if not, skip it.
            # Possible reasons are e.g: No access (via GitHub app or Gerrit user),
            # Clone/checkout failures for plain git repos or similar.
            if not repo._repo:
                LOGGER.error(
                    "Repo '%s' could not be initialized. Skip scraping.", repo_name
                )
                continue

            # Build the data for the repo itself to be stored in Elasticsearch
            uuid = hashlib.sha1(str.encode(repo_name)).hexdigest()
            es_repo = GitRepo(meta={"id": uuid})
            es_repo.repo_name = repo_name
            es_repo.scrape_time = scrape_time
            es_repo.provider = provider
            es_repos.append(es_repo)

            # scrape the repo if is part of the tenant config
            scrape_repo(repo, tenants, scrape_time)

        # Store the information for all repos we just scraped in Elasticsearch
        LOGGER.info("Updating %d repo definitions in Elasticsearch", len(es_repos))
        GitRepo.bulk_save(es_repos)
    else:
        # Delete the repositories from the repo_cache
        for repo_name in repo_list:
            repo_cache.pop(repo_name, None)

    # In both cases we want to delete outdated data.
    # In case of delete_only, this will be everything!
    # NOTE (felix): In case of a config error, the repo is removed from this list
    LOGGER.info("Deleting outdated data for the following repositories: %s", repo_list)
    delete_outdated(
        scrape_time, [AnsibleRole, ZuulJob], extra_filter=Q("terms", repo=repo_list)
    )

    LOGGER.info("Deleting the following repositories (only if outdated): %s", repo_list)
    # NOTE (fschmidt): Usually, this should not delete anything we just scraped.
    delete_outdated(
        scrape_time,
        [GitRepo],
        extra_filter=Q({"terms": {"repo_name.keyword": repo_list}}),
    )


def scrape_repo(repo, tenants, scrape_time):
    job_files, role_files = Scraper(repo).scrape()

    jobs, roles = RepoParser(repo, tenants, job_files, role_files, scrape_time).parse()

    LOGGER.info("Updating %d job definitions in Elasticsearch", len(jobs))
    ZuulJob.bulk_save(jobs)

    LOGGER.info("Updating %d role definitions in Elasticsearch", len(roles))
    AnsibleRole.bulk_save(roles)


def delete_outdated(scrape_time, indices, extra_filter=None):
    # Delete all outdated entries in Elasticsearch
    LOGGER.info(
        "Going to delete outdated scraping results which are older than %s",
        datetime.strftime(scrape_time, "%Y-%m-%dT%H:%M:%SZ"),
    )
    for index in indices:
        try:
            deleted_docs = index.outdated_query(scrape_time, extra_filter).delete()
            LOGGER.info(
                "Deleted %d outdated %ss in Elasticsearch",
                deleted_docs.deleted,
                index.__name__,
            )
        except ConflictError:
            LOGGER.info("Deleted 0 outdated %ss in Elasticsearch", index.__name__)


# TODO (fschmidt): Maybe it's worth to move the event_* methods to a GitHubEventHandler
# class or similar. This way, we could encapsulate different events in their respective
# environment (e.g. GitHub, Gerrit, ...)
def handle_event(event, payload, config, connections, tenant_parser, repo_cache):
    LOGGER.info("Handling event '%s'", event)
    try:
        # TODO (fschmidt): Maybe we should change this file/module to be a class
        # to get rid of this module lookup
        this_module = sys.modules[__name__]
        method = getattr(this_module, "event_{}".format(event))
    except AttributeError:
        LOGGER.warning(
            "Could not find an appropriate method to handle event '%s'", event
        )
        return

    try:
        # TODO (fschmidt): What about 'repository' events?
        # To get updates for public/private?
        # https://developer.github.com/v3/activity/events/types/#repositoryevent
        method(payload, config, connections, tenant_parser, repo_cache)
    except Exception:
        # TODO (fschmidt): Does it make sense to catch an Exception here?
        # Could we catch anything more specific?
        LOGGER.exception("Error while handling event '%s'", event)


def event_installation(payload, config, connections, tenant_parser, repo_cache):
    action = payload.get("action")
    installation_id = payload.get("installation", {}).get("id")
    repositories = payload.get("repositories", [])

    if action == "created":
        LOGGER.info("Scraping new installation %d", installation_id)
        # Get list of repos from the payload
        repo_names = [r["full_name"] for r in repositories]
        # Scrape them
        scrape_repo_list(repo_names, connections, tenant_parser, repo_cache=repo_cache)

    if action == "deleted":
        LOGGER.info("Deleting data for installation %d", installation_id)
        # TODO (felix) Get the right connection from the configuration based on what?
        # The provider? The github url? Both?
        gh_con = connections["github"]
        # Get repos for this installation from our GitHubConnection as they are
        # not listed in the payload.
        # FIXME (fschmidt): When the repo map is updated between triggering and
        # handling the event, the repo_list can no longer be retrieved from
        # the installation (which is already deleted by then).
        # As a fallback, we could search for all repos in ES which start with
        # the organization (<orga>/*)
        repositories = gh_con.get_repos_for_installation(installation_id)
        if not repositories:
            LOGGER.warning(
                "Could not retrieve repo list for installation %d. "
                "Maybe we don't have access any longer.",
                installation_id,
            )
            LOGGER.warning("Nothing to delete")
            return
        # Delete all data for those repos
        scrape_repo_list(
            repositories,
            connections,
            tenant_parser,
            repo_cache=repo_cache,
            delete_only=True,
        )

        # TODO (fschmidt): Should we remove them also from the installatino map?


def event_installation_repositories(
    payload, config, connections, tenant_parser, repo_cache
):
    installation_id = payload.get("installation", {}).get("id")
    repos_added = payload.get("repositories_added")
    repos_removed = payload.get("repositories_removed")

    project_name = payload.get("repository", {}).get("full_name")
    print(installation_id)
    print(project_name)

    # TODO validate installation id from payload against installation map?
    # If they do not match, we might have missed an installation event and
    # should update our installation map

    # Scrape each added repo
    if repos_added is not None:
        LOGGER.info(
            "Scraping %d new repositories for installation %d",
            len(repos_added),
            installation_id,
        )
        # Get list of repos from the payload
        repo_names = [r["full_name"] for r in repos_added]
        # Scrape them
        scrape_repo_list(repo_names, connections, tenant_parser, repo_cache=repo_cache)

    # Just delete the data for these repos
    if repos_removed is not None:
        LOGGER.info(
            "Deleting data from %d repositories for installation %d",
            len(repos_removed),
            installation_id,
        )
        # Get list of repos from the payload
        repo_names = [r["full_name"] for r in repos_removed]
        # Delete all data for those repos
        scrape_repo_list(
            repo_names,
            connections,
            tenant_parser,
            repo_cache=repo_cache,
            delete_only=True,
        )


def event_push(payload, config, connections, tenant_parser, repo_cache):
    repo_name = payload.get("repository", {}).get("full_name")
    # installation_id = payload.get('installation', {}).get('id')
    ref = payload.get("ref")
    # old_ref = payload.get('before')
    # new_ref = payload.get('after')
    # commits = payload.get('commits')

    # TODO (felix) Get the right connection from the configuration based on what?
    # The provider? The github url? Both?
    gh_con = connections["github"]

    repo_info = gh_con.installation_map.get(repo_name)
    default_branch = repo_info["default_branch"]

    # TODO validate installation id from payload against installation map?
    # If they do not match, we might have missed an installation event and
    # should update our installation map

    parts = ref.split("/", 2)
    branch = parts[2]
    if branch != default_branch:
        LOGGER.info(
            "Push event contains ref %s, but default branch is %s. "
            "Won't handle event for repo %s.",
            ref,
            default_branch,
            repo_name,
        )
        return

    LOGGER.info("Handling push event for repo %s with ref %s", repo_name, ref)

    scrape_repo_list([repo_name], connections, tenant_parser, repo_cache=repo_cache)


if __name__ == "__main__":
    main()



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class RepositoryError(Exception):
    """Base class for all repository related errors."""

    pass


class CheckoutError(RepositoryError):
    """Exception if a file could not be checked out."""

    def __init__(self, path, cause, msg=None):
        if msg is None:
            msg = "Failed to check out '{}': {}".format(path, cause)
        super().__init__(msg)


class ScraperConfigurationError(Exception):
    """Exception if the scraper configuration is wrong."""

    pass



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

import yaml

from zubbi.scraper.exceptions import CheckoutError, ScraperConfigurationError


TENANTS_DIRECTORY = "tenants"

LOGGER = logging.getLogger(__name__)


class TenantParser:
    def __init__(self, sources_file=None, sources_repo=None):
        self.sources = None
        self.repo_map = {}
        self.tenants = []
        # Initial call to load the sources file/repo
        self.reload_sources(sources_file, sources_repo)

    def reload_sources(self, sources_file=None, sources_repo=None):
        if sources_file:
            self.sources = self._load_tenant_sources_from_file(sources_file)
        else:
            self.sources = self._load_tenant_sources_from_repo(sources_repo)

    def update(self, sources_file=None, sources_repo=None):
        self.reload_sources(sources_file, sources_repo)
        self.parse()

    def parse(self):
        # Clear repo_map and tenant list first
        self.repo_map.clear()
        self.tenants.clear()

        for tenant_src in self.sources:
            sources = tenant_src["tenant"]["source"]
            tenant_name = tenant_src["tenant"]["name"]

            # Iterate over all repositories specified in this tenant (per provider)
            for connection_name, source in sources.items():
                # project_type is config- or untrusted-project
                for project_type, projects in source.items():
                    for project in projects:
                        self._update_repo_map(project, connection_name, tenant_name)

            self.tenants.append(tenant_name)

    def _update_repo_map(self, project, connection_name, tenant):
        project_name, exclude = self._extract_project(project)

        # Map the current tenant to the current repository
        repo_tenant_entry = self.repo_map.setdefault(
            project_name,
            {"tenants": {"jobs": [], "roles": []}, "connection_name": connection_name},
        )

        # Update repo_tenant mapping
        if "jobs" not in exclude:
            repo_tenant_entry["tenants"]["jobs"].append(tenant)
        repo_tenant_entry["tenants"]["roles"].append(tenant)

    def _extract_project(self, project):
        project_name = project
        exclude = []
        if type(project) is dict:
            # Get the first key of the dict containing the project name.
            project_name = list(project.keys())[0]
            exclude = project.get("exclude", [])
        return project_name, exclude

    def _load_tenant_sources_from_file(self, sources_file):
        LOGGER.info("Parsing tenant sources file '%s'", sources_file)
        with open(sources_file) as f:
            sources = yaml.load(f)
        return sources

    def _load_tenant_sources_from_repo(self, sources_repo):
        LOGGER.info("Collecting tenant sources from repo '%s'", sources_repo)
        sources = []
        try:
            tenants = sources_repo.list_directory(TENANTS_DIRECTORY)
        except CheckoutError:
            raise ScraperConfigurationError(
                "Cannot load tenant sources. Repo '%s' does not contain a "
                "'tenants' folder".format(sources_repo.repo_name)
            )

        for tenant in tenants.keys():
            try:
                sources_yaml = sources_repo.check_out_file(
                    os.path.join("tenants", tenant, "sources.yaml")
                )
                settings_yaml = sources_repo.check_out_file(
                    os.path.join("tenants", tenant, "settings.yaml")
                )
                # NOTE (fschmidt): We parse both files and create the same data
                # structure like zuul does for the main.yaml file.
                tenant_sources = {
                    # Load the settings first, as they contain different keys
                    "tenant": yaml.load(settings_yaml)
                }
                # Update the tenant_sources with the sources file and wrap them
                # in a 'source' key
                tenant_sources["tenant"]["source"] = yaml.load(sources_yaml)

                sources.append(tenant_sources)
            except CheckoutError as e:
                # If a single tenant is missing the required file, we just skip it
                LOGGER.warning(
                    "Either 'settings.yaml' or 'sources.yaml' are "
                    "missing or empty in repo '%s': %s",
                    sources_repo.repo_name,
                    e,
                )
        return sources



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from flask import Blueprint, Flask

from . import default_settings
from .helpers import init_helpers, make_error_handler
from .models import init_elasticsearch
from .views import (
    AutoCompleteView,
    DetailView,
    HowToView,
    IndexView,
    LmzifyView,
    SearchView,
    WebhookView,
)


class ZubbiBlueprint(Blueprint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        IndexView.register_url(self)
        SearchView.register_url(self)
        DetailView.register_url(self)
        HowToView.register_url(self)
        AutoCompleteView.register_url(self)
        WebhookView.register_url(self)
        LmzifyView.register_url(self)

        self.record(self._init_blueprint)
        self.record_once(self._init_blueprint_once)

    def _init_blueprint(self, state):
        for key in dir(default_settings):
            if key.isupper():
                state.app.config.setdefault(key, getattr(default_settings, key))
        init_helpers(state.app)

    def _init_blueprint_once(self, state):
        init_elasticsearch(state.app)


blueprint = ZubbiBlueprint(
    "zubbi", __name__, template_folder="templates", static_folder="static"
)


ZUBBI_SETTINGS_ENV = "ZUBBI_SETTINGS"


def create_app(config=None):
    app = Flask(__name__)

    if config is None:
        app.config.from_envvar(ZUBBI_SETTINGS_ENV)
    else:
        app.config.from_mapping(config)

    app.register_blueprint(blueprint)

    error_handler = make_error_handler()

    app.register_error_handler(400, error_handler)
    app.register_error_handler(404, error_handler)

    try:
        os.makedirs(app.instance_path)
    except FileExistsError:
        pass

    return app



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_packages, setup


description = "Index for finding jobs & roles used in a Zuul based CI system"


with open("README.md") as readme:
    long_description = readme.read()


requires = [
    "arrow",
    "click",
    "elasticsearch-dsl",
    "pyyaml",
    "pyjwt",
    "pyzmq",
    "cryptography",
    "requests",
    "flask",
    "github3.py",
    "zuul-sphinx",
    "sphinx",
    "jinja2",
    "readme_renderer[md]",
    "tabulate",
    "gitpython",
]

setup(
    name="zubbi",
    description=description,
    author="Benedikt Loeffler, Felix Schmidt, Simon Westphahl",
    author_email="benedikt.loeffler@bmw.de, felix.fs.schmidt@bmw.de, "
    "simon.westphahl@bmw.de",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
    install_requires=requires,
    url="https://github.com/bmwcarit/zubbi",
    license="Apache License 2.0",
    use_scm_version={"version_scheme": "post-release"},
    setup_requires=["setuptools_scm"],
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    entry_points={
        "console_scripts": ["zubbi-scraper = zubbi.scraper.main:main"],
        "flask.commands": ["collectstatic=zubbi.cli:collectstatic"],
    },
    classifiers={
        "Development Status :: 5 - Production/Stable",
        "Framework :: Flask",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3 :: Only",
    },
)
